import warnings
import aerosandbox as asb
import aerosandbox.numpy as np
from dataclasses import dataclass, field
from typing import Union, Sequence, List, Any, Dict
from scipy import interpolate


def compute_optimal_x_points(
        n_points,
):
    # theta = np.arccos(1 - 2 * x)
    # delta_Cp_thin_airfoil_theory = 4 / np.tan(theta)
    # desired_spacing = 1 / delta_Cp_thin_airfoil_theory
    # spacing_function = integral(desired_spacing, x) # rescaled to (0, 1)

    # thin_airfoil_theory_spacing_function = lambda x: (
    #                                                          -np.sqrt((1 - x) * x)
    #                                                          - 0.5 * np.arcsin(1 - 2 * x)
    #                                                  ) / (np.pi / 2) + 0.5
    #
    # uniform_spacing_function = lambda x: x
    #
    # return (
    #         thin_airfoil_theory_spacing_function(np.linspace(0, 1, n_points)) +
    #         uniform_spacing_function(np.sinspace(0, 1, n_points))
    # ) / 2
    s = np.linspace(0, 1, n_points + 1)
    return (s[1:] + s[:-1]) / 2


# def compute_optimal_x_points(
#         n_points: int = 24,
# ):
#     # theta = np.arccos(1 - 2 * x)
#     # delta_Cp_thin_airfoil_theory = 4 / np.tan(theta)
#     # desired_spacing = 1 / delta_Cp_thin_airfoil_theory
#     # spacing_function = integral(desired_spacing, x) # rescaled to (0, 1)
#
#     thin_airfoil_theory_spacing_function = lambda x: (
#                                                              -np.sqrt((1 - x) * x)
#                                                              - 0.5 * np.arcsin(1 - 2 * x)
#                                                      ) / (np.pi / 2) + 0.5
#
#     uniform_spacing_function = lambda x: x
#
#     return (
#             thin_airfoil_theory_spacing_function(np.linspace(0, 1, n_points)) +
#             uniform_spacing_function(np.sinspace(0, 1, n_points))
#     ) / 2


@dataclass
class Data():
    airfoil: asb.KulfanAirfoil
    alpha: float
    Re: float
    mach: float
    n_crit: float
    xtr_upper: float
    xtr_lower: float

    N = 32
    bl_x_points = compute_optimal_x_points(n_points=N)

    analysis_confidence: float  # Nominally 0 (no confidence) to 1 (high confidence)
    af_outputs: Dict[str, Any] = field(
        default_factory=lambda: {
            "CL"     : np.nan,
            "CD"     : np.nan,
            "CM"     : np.nan,
            "Top_Xtr": np.nan,
            "Bot_Xtr": np.nan,
        }
    )
    upper_bl_outputs: Dict[str, Any] = field(
        default_factory=lambda: {
            "theta"  : np.nan * np.ones_like(Data.bl_x_points),
            "H"      : np.nan * np.ones_like(Data.bl_x_points),
            "ue/vinf": np.nan * np.ones_like(Data.bl_x_points),
        }
    )  # theta, H, ue/vinf
    lower_bl_outputs: Dict[str, Any] = field(
        default_factory=lambda: {
            "theta"  : np.nan * np.ones_like(Data.bl_x_points),
            "H"      : np.nan * np.ones_like(Data.bl_x_points),
            "ue/vinf": np.nan * np.ones_like(Data.bl_x_points),
        }
    )  # theta, H, ue/vinf

    @property
    def inputs(self):
        return {
            "airfoil"  : self.airfoil,
            "alpha"    : self.alpha,
            "Re"       : self.Re,
            "mach"     : self.mach,
            "n_crit"   : self.n_crit,
            "xtr_upper": self.xtr_upper,
            "xtr_lower": self.xtr_lower,
        }

    @property
    def outputs(self):
        return {
            "analysis_confidence": self.analysis_confidence,
            "af_outputs"         : self.af_outputs,
            "upper_bl_outputs"   : self.upper_bl_outputs,
            "lower_bl_outputs"   : self.lower_bl_outputs,
        }

    @classmethod
    def from_xfoil(cls,
                   airfoil: Union[asb.Airfoil, asb.KulfanAirfoil],
                   alphas: Union[float, Sequence[float]],
                   Re: float,
                   mach: float,
                   n_crit: float,
                   xtr_upper: float,
                   xtr_lower: float,
                   timeout=5,
                   max_iter=100,
                   xfoil_command: str = 'xfoil'
                   ) -> List["Data"]:
        airfoil = airfoil.normalize().to_kulfan_airfoil()

        alphas = np.atleast_1d(alphas)

        xf = asb.XFoil(
            airfoil=airfoil,
            Re=Re,
            mach=mach,
            n_crit=n_crit,
            xtr_upper=xtr_upper,
            xtr_lower=xtr_lower,
            xfoil_repanel=True,
            timeout=timeout,
            max_iter=max_iter,
            xfoil_command=xfoil_command,
            include_bl_data=True,
        )
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=UserWarning)
            xf_outputs = xf.alpha(alphas)

        training_datas = []

        def append_empty_data():
            training_datas.append(
                cls(
                    airfoil=airfoil.to_kulfan_airfoil(),
                    alpha=alpha,
                    Re=Re,
                    mach=mach,
                    n_crit=n_crit,
                    xtr_upper=xtr_upper,
                    xtr_lower=xtr_lower,
                    analysis_confidence=0,
                )
            )

        for i, alpha in enumerate(alphas):
            ### Figure out which output corresponds to this alpha
            alpha_deviations = np.abs(xf_outputs["alpha"] - alpha)

            if (  # If the alpha is not in the output
                    (len(alpha_deviations) == 0) or
                    (np.min(alpha_deviations) > 0.001)
            ):
                append_empty_data()
                continue

            index = np.argmin(alpha_deviations)
            xf_output = {
                key: value[index] for key, value in xf_outputs.items()
            }
            bl_data = xf_output["bl_data"]

            # ### Trim off the wake data (rows where "Cp" first becomes NaN)
            # wake_node_indices = np.flatnonzero(bl_data["x"] > )
            # bl_data = bl_data.iloc[
            #           :np.flatnonzero(bl_data["x"] <= airfoil.x().max())[-1] + 1,
            #           :
            #           ]

            ### Split the boundary layer data into upper and lower sections
            # LE_index = bl_data["x"].argmin()
            try:
                dx = np.diff(bl_data["x"].values)
                upper_bl_data = bl_data.iloc[
                                :np.flatnonzero(dx < 0)[-1] + 2
                                :
                                ].iloc[::-1, :]
                lower_bl_data = bl_data.iloc[
                                np.flatnonzero(dx > 0)[0]:,
                                :
                                ]
            except IndexError as e:  # If the boundary layer data is too short
                print(e)
                append_empty_data()
                continue

            if len(upper_bl_data) <= 4 or len(lower_bl_data) <= 4:  # If the boundary layer data is too short
                print("BL data too short")
                append_empty_data()
                continue

            interp = lambda x, y: interpolate.PchipInterpolator(x, y, extrapolate=True)(cls.bl_x_points)

            try:
                training_datas.append(
                    cls(
                        airfoil=airfoil.to_kulfan_airfoil(),
                        alpha=alpha,
                        Re=Re,
                        mach=mach,
                        n_crit=n_crit,
                        xtr_upper=xtr_upper,
                        xtr_lower=xtr_lower,
                        analysis_confidence=1,
                        af_outputs={
                            "CL"     : xf_output["CL"],
                            "CD"     : xf_output["CD"],
                            "CM"     : xf_output["CM"],
                            "Top_Xtr": xf_output["Top_Xtr"],
                            "Bot_Xtr": xf_output["Bot_Xtr"],
                        },
                        upper_bl_outputs={
                            "theta"  : interp(upper_bl_data["x"], upper_bl_data["theta"]),
                            "H"      : interp(upper_bl_data["x"], upper_bl_data["H"]),
                            "ue/vinf": interp(upper_bl_data["x"], upper_bl_data["ue/vinf"]),
                        },
                        lower_bl_outputs={
                            "theta"  : interp(lower_bl_data["x"], lower_bl_data["theta"]),
                            "H"      : interp(lower_bl_data["x"], lower_bl_data["H"]),
                            "ue/vinf": interp(lower_bl_data["x"], lower_bl_data["ue/vinf"]),
                        },
                    )
                )
            except ValueError as e:
                print(e)
                append_empty_data()
                continue

        return training_datas

    def to_vector(self) -> np.ndarray:  # dtype: float32
        items = [
            self.airfoil.upper_weights,
            self.airfoil.lower_weights,
            self.airfoil.leading_edge_weight,
            self.airfoil.TE_thickness,
            self.alpha,
            self.Re,
            self.mach,
            self.n_crit,
            self.xtr_upper,
            self.xtr_lower,
            self.analysis_confidence,
            self.af_outputs["CL"],
            self.af_outputs["CD"],
            self.af_outputs["CM"],
            self.af_outputs["Top_Xtr"],
            self.af_outputs["Bot_Xtr"],
            self.upper_bl_outputs["theta"],
            self.upper_bl_outputs["H"],
            self.upper_bl_outputs["ue/vinf"],
            self.lower_bl_outputs["theta"],
            self.lower_bl_outputs["H"],
            self.lower_bl_outputs["ue/vinf"],
        ]

        return np.concatenate([
            np.atleast_1d(item).flatten().astype(np.float32)
            for item in items
        ], axis=0)

    @classmethod
    def from_vector(cls, vector: np.ndarray) -> "Data":
        i = 0

        def pop(n):
            nonlocal i
            result = vector[i:i + n]
            i += n
            return result

        airfoil = asb.KulfanAirfoil(
            upper_weights=pop(8),
            lower_weights=pop(8),
            leading_edge_weight=pop(1),
            TE_thickness=pop(1),
        )
        alpha = pop(1)
        Re = pop(1)
        mach = pop(1)
        n_crit = pop(1)
        xtr_upper = pop(1)
        xtr_lower = pop(1)
        analysis_confidence = pop(1)
        af_outputs = {
            "CL"     : pop(1),
            "CD"     : pop(1),
            "CM"     : pop(1),
            "Top_Xtr": pop(1),
            "Bot_Xtr": pop(1),
        }
        upper_bl_outputs = {
            "theta"  : pop(cls.N),
            "H"      : pop(cls.N),
            "ue/vinf": pop(cls.N),
        }
        lower_bl_outputs = {
            "theta"  : pop(cls.N),
            "H"      : pop(cls.N),
            "ue/vinf": pop(cls.N),
        }
        if not i == len(vector):
            raise ValueError(f"Vector length mismatch: Got {len(vector)}, expected {i}.")

        return cls(
            airfoil=airfoil,
            alpha=alpha,
            Re=Re,
            mach=mach,
            n_crit=n_crit,
            xtr_upper=xtr_upper,
            xtr_lower=xtr_lower,
            analysis_confidence=analysis_confidence,
            af_outputs=af_outputs,
            upper_bl_outputs=upper_bl_outputs,
            lower_bl_outputs=lower_bl_outputs,
        )

    @classmethod
    def get_vector_input_column_names(cls):
        return [
            *[f"kulfan_upper_{i}" for i in range(8)],
            *[f"kulfan_lower_{i}" for i in range(8)],
            "kulfan_LE_weight",
            "kulfan_TE_thickness",
            "alpha",
            "Re",
            "mach",
            "n_crit",
            "xtr_upper",
            "xtr_lower",
        ]

    @classmethod
    def get_vector_output_column_names(cls):
        return [
            "analysis_confidence",
            "CL",
            "CD",
            "CM",
            "Top_Xtr",
            "Bot_Xtr",
            *[f"upper_bl_theta_{i}" for i in range(cls.N)],
            *[f"upper_bl_H_{i}" for i in range(cls.N)],
            *[f"upper_bl_ue/vinf_{i}" for i in range(cls.N)],
            *[f"lower_bl_theta_{i}" for i in range(cls.N)],
            *[f"lower_bl_H_{i}" for i in range(cls.N)],
            *[f"lower_bl_ue/vinf_{i}" for i in range(cls.N)],
        ]

    @classmethod
    def get_vector_column_names(cls):
        return cls.get_vector_input_column_names() + cls.get_vector_output_column_names()

    def __eq__(self, other: "Data") -> bool:
        v1 = self.to_vector()
        v2 = other.to_vector()

        return all([
            np.allclose(
                s1, s2,
                atol=0,
                rtol=np.finfo(np.float32).eps * 10,
            ) or np.all(np.isnan([s1, s2]))
            for s1, s2 in zip(v1, v2)
        ])

        return np.allclose(
            self.to_vector(),
            other.to_vector(),
            atol=0,
            rtol=np.finfo(np.float32).eps * 10,
        )

    def validate_vector_format(self):
        assert self == Data.from_vector(self.to_vector())


if __name__ == '__main__':
    af = asb.Airfoil("ag41d-02r").normalize().to_kulfan_airfoil()

    datas = Data.from_xfoil(
        airfoil=af,
        alphas=[3, 5, 60],
        Re=1e6,
        mach=0,
        n_crit=9,
        xtr_upper=0.99,
        xtr_lower=0.99,
    )

    d = datas[0]

    # import aerosandbox as asb
    # import aerosandbox.numpy as np
    #
    # airfoil_database_path = asb._asb_root / "geometry" / "airfoil" / "airfoil_database"
    #
    # UIUC_airfoils = [
    #     asb.Airfoil(name=filename.stem).normalize().to_kulfan_airfoil()
    #     for filename in airfoil_database_path.glob("*.dat")
    # ]
    #
    # for af in UIUC_airfoils:
    #     print(af.name)
    #     datas = Data.from_xfoil(
    #         airfoil=af,
    #         alphas=[2, 60],
    #         Re=1e6,
    #         mach=0,
    #         n_crit=9,
    #         xtr_upper=0.99,
    #         xtr_lower=0.99,
    #     )
    #
    #     d = datas[0]
    #     # print(d.to_vector())
    #     # print(Data.from_vector(d.to_vector()))
    #     print(d == Data.from_vector(d.to_vector()))



import aerosandbox as asb
import aerosandbox.numpy as np
from typing import Union, Dict, Set, List
from pathlib import Path

npz_file_directory = Path(__file__).parent / "nn_weights_and_biases"


def get_aero_from_kulfan_parameters(
        kulfan_parameters: Dict[str, Union[float, np.ndarray]],
        alpha: Union[float, np.ndarray],
        Re: Union[float, np.ndarray],
        model_size="large"
) -> Dict[str, Union[float, np.ndarray]]:

    ### Load the neural network parameters
    filename = npz_file_directory / f"nn-{model_size}.npz"
    if not filename.exists():
        raise FileNotFoundError(
            f"Could not find the neural network file {filename}, which contains the weights and biases.")

    data: Dict[str, np.ndarray] = np.load(filename)

    ### Prepare the inputs for the neural network
    input_rows: List[Union[float, np.ndarray]] = [
        4 * np.sind(2 * alpha),
        20 * (1 - np.cosd(alpha) ** 2),
        (np.log(Re) - 12.5) / 2,
        *[kulfan_parameters["lower_weights"][i] * 5 for i in range(8)],
        *[kulfan_parameters["upper_weights"][i] * 5 for i in range(8)],
        kulfan_parameters["leading_edge_weight"] * 5,
        kulfan_parameters["TE_thickness"] * 100,
    ]

    ### Handle the vectorization, where here we figure out how many cases the user wants to run
    N_cases = 1  # TODO rework this with np.atleast1d
    for row in input_rows:
        if np.length(row) > 1:
            if N_cases == 1:
                N_cases = np.length(row)
            else:
                if np.length(row) != N_cases:
                    raise ValueError(
                        f"The inputs to the neural network must all have the same length. (Conflicting lengths: {N_cases} and {np.length(row)})"
                    )

    for i, row in enumerate(input_rows):
        input_rows[i] = np.ones(N_cases) * row

    x = np.stack(input_rows, axis=0)

    ##### Evaluate the neural network

    ### First, determine what the structure of the neural network is (i.e., how many layers it has) by looking at the keys.
    # These keys come from the dictionary of saved weights/biases for the specified neural network.
    try:
        layer_indices: Set[int] = set([
            int(key.split(".")[1])
            for key in data.keys()
        ])
    except TypeError:
        raise ValueError(
            f"Got an unexpected neural network file format.\n"
            f"Dictionary keys should be strings of the form 'net.0.weight', 'net.0.bias', 'net.2.weight', etc.'.\n"
            f"Instead, got keys of the form {data.keys()}.\n"
        )
    layer_indices: List[int] = sorted(list(layer_indices))

    ### Now, set up evaluation of the basic neural network.
    def net(x: np.ndarray):
        layer_indices_to_iterate = layer_indices.copy()

        while len(layer_indices_to_iterate) != 0:
            i = layer_indices_to_iterate.pop(0)
            w = data[f"net.{i}.weight"]
            b = data[f"net.{i}.bias"]
            x = w @ x + np.reshape(b, (-1, 1))

            if len(layer_indices_to_iterate) != 0:  # Don't apply the activation function on the last layer
                x = np.tanh(x)

        return x

    y = net(x)

    ### Then, flip the inputs and evaluate the network again.
    # The goal here is to embed the invariant of "symmetry across alpha" into the network evaluation.
    # (This was also performed during training, so the network is "intended" to be evaluated this way.)
    input_rows_flipped: List[Union[float, np.ndarray]] = [
        -4 * np.sind(2 * alpha),
        20 * (1 - np.cosd(alpha) ** 2),
        (np.log(Re) - 12.5) / 2,
        *[kulfan_parameters["upper_weights"][i] * -5 for i in range(8)],
        *[kulfan_parameters["lower_weights"][i] * -5 for i in range(8)],
        kulfan_parameters["leading_edge_weight"] * -5,
        kulfan_parameters["TE_thickness"] * 100,
    ]

    for i, row in enumerate(input_rows_flipped):
        input_rows_flipped[i] = np.ones(N_cases) * row

    x_flipped = np.stack(input_rows_flipped, axis=0)

    y_flipped = net(x_flipped)

    ### The resulting outputs will also be flipped, so we need to flip them back to their normal orientation
    y_flipped[0, :] *= -1  # CL
    y_flipped[2, :] *= -1  # CM
    temp = y_flipped[4, :] + 0.  # This is here to help swap the top / bottom Xtr (transition x) values
    # Adding the 0. is a hack to force a memory-copy of the array to be made in a way that's array-backend-agnostic.
    y_flipped[4, :] = y_flipped[5,:]  # Replace top Xtr with bottom Xtr
    y_flipped[5, :] = temp  # Replace bottom Xtr with top Xtr

    ### Then, average the two outputs to get the "symmetric" result
    y = (y + y_flipped) / 2

    # Unpack the neural network outputs
    results = {
        "CL"     : y[0, :],
        "CD"     : np.exp(y[1, :] - 4),
        "CM"     : y[2, :] / 20,
        "Cpmin"  : 1 - y[3, :] ** 2,
        "Top_Xtr": y[4, :],
        "Bot_Xtr": y[5, :],
    }
    return {key: np.reshape(value, -1) for key, value in results.items()}


def get_aero_from_airfoil(
        airfoil: asb.Airfoil,
        alpha: Union[float, np.ndarray],
        Re: Union[float, np.ndarray],
        model_size="large",
) -> Dict[str, Union[float, np.ndarray]]:

    airfoil_normalization = airfoil.normalize(return_dict=True)

    from aerosandbox.geometry.airfoil.airfoil_families import get_kulfan_parameters

    kulfan_parameters = get_kulfan_parameters(
        airfoil_normalization["airfoil"].coordinates,
        n_weights_per_side=8
    )

    return get_aero_from_kulfan_parameters(
        kulfan_parameters=kulfan_parameters,
        alpha=alpha + airfoil_normalization["rotation_angle"],
        Re=Re / airfoil_normalization["scale_factor"],
        model_size=model_size
    )


def get_aero_from_coordinates(
        coordinates: np.ndarray,
        alpha: Union[float, np.ndarray],
        Re: Union[float, np.ndarray],
        model_size="large",
):
    return get_aero_from_airfoil(
        airfoil=asb.Airfoil(
            coordinates=coordinates
        ),
        alpha=alpha,
        Re=Re,
        model_size=model_size
    )


def get_aero_from_dat_file(
        filename,
        alpha: Union[float, np.ndarray],
        Re: Union[float, np.ndarray],
        model_size="large",
):
    with open(filename, "r") as f:
        raw_text = f.readlines()

    from aerosandbox.geometry.airfoil.airfoil_families import get_coordinates_from_raw_dat
    return get_aero_from_coordinates(
        coordinates=get_coordinates_from_raw_dat(raw_text=raw_text),
        alpha=alpha,
        Re=Re,
        model_size=model_size
    )


if __name__ == '__main__':

    airfoil = asb.Airfoil("dae11").repanel().normalize()
    # airfoil = asb.Airfoil("naca0012").add_control_surface(10, hinge_point_x=0.5)

    alpha = np.linspace(-15, 15, 100)
    Re = 1e6

    aeros = {}

    model_sizes = ["xxsmall", "xsmall", "small", "medium", "large", "xlarge", "xxlarge", "xxxlarge"]

    for model_size in model_sizes:
        aeros[f"NF-{model_size}"] = get_aero_from_airfoil(
            airfoil=airfoil,
            alpha=alpha,
            Re=Re,
            model_size=model_size
        )

    if False:

        aeros["XFoil"] = asb.XFoil(
            airfoil=airfoil,
            Re=Re,
            max_iter=20,
            xfoil_repanel=True
        ).alpha(alpha)

    import matplotlib.pyplot as plt
    import aerosandbox.tools.pretty_plots as p

    fig, ax = plt.subplots(2, 2, figsize=(10, 8))
    for label, aero in aeros.items():
        if "xfoil" in label.lower():
            a = aero["alpha"]
            kwargs = dict(
                color="k"
            )
        else:
            a = alpha
            kwargs = dict(
                alpha=0.6,
            )

        ax[0, 0].plot(a, aero["CL"], **kwargs)
        ax[0, 1].plot(a, aero["CD"], label=label, **kwargs)
        ax[1, 0].plot(a, aero["CM"], **kwargs)
        ax[1, 1].plot(aero["CD"], aero["CL"], **kwargs)
    plt.sca(ax[0, 1])
    plt.legend()
    ax[0, 0].set_xlabel("$\\alpha$")
    ax[0, 0].set_ylabel("$C_L$")

    ax[0, 1].set_xlabel("$\\alpha$")
    ax[0, 1].set_ylabel("$C_D$")
    ax[0, 1].set_ylim(bottom=0)

    ax[1, 0].set_xlabel("$\\alpha$")
    ax[1, 0].set_ylabel("$C_M$")

    ax[1, 1].set_xlabel("$C_D$")
    ax[1, 1].set_ylabel("$C_L$")
    ax[1, 1].set_xlim(left=0)

    from aerosandbox.tools.string_formatting import eng_string

    plt.suptitle(f"\"{airfoil.name}\" Airfoil at $Re_c = \\mathrm{{{eng_string(Re)}}}$")

    p.show_plot()



import pytest
import neuralfoil as nf
import aerosandbox as asb
import aerosandbox.numpy as np

def test_basic_functionality():
    aero = nf.get_aero_from_airfoil(
        asb.Airfoil("naca4412"),
        alpha=5,
        Re=1e6
    )

if __name__ == '__main__':
    pytest.main()


import aerosandbox as asb
import aerosandbox.numpy as np
from typing import Union


def get_CL(
        airfoil: asb.Airfoil,
        alpha: Union[float, np.ndarray],
        Re: Union[float, np.ndarray],
) -> Union[float, np.ndarray]:
    p = {
        'CLa_low'            : 5.796551904649262,
        'CLa_high'           : 6.662038127483676,
        'log10_Re_switch'    : 4.759293563062047,
        'log10_Re_scale'     : 0.40489480755424206,
        'CL0_high_LE'        : 0.17280052131807694,
        'CL0_high_0'         : 0.31799245171528784,
        'CL0_high_1'         : 0.2826151876712608,
        'CL0_high_2'         : 0.24841615861002925,
        'CL0_high_3'         : 0.24016960597766465,
        'CL0_high_4'         : 0.23497448970552062,
        'CL0_high_5'         : 0.23179443674621486,
        'CL0_high_6'         : 0.27764987935170676,
        'CL0_high_7'         : 0.3558323531831959,
        'CL0_low_LE'         : 0.12160802332276614,
        'CL0_low_0'          : 0.11754623987319483,
        'CL0_low_1'          : 0.038035213950785456,
        'CL0_low_2'          : -0.1108768707653387,
        'CL0_low_3'          : -0.10011194827235753,
        'CL0_low_4'          : -0.019534698670369936,
        'CL0_low_5'          : 0.057508354923867314,
        'CL0_low_6'          : 0.10926620536707825,
        'CL0_low_7'          : -0.10812301223195821,
        'delta_CLa_high_area': -0.2887303436003777,
        'delta_CLa_low_area' : -1.2434390140977036
    }

    from aerosandbox.geometry.airfoil.airfoil_families import get_kulfan_parameters
    kulfan_params = get_kulfan_parameters(airfoil.coordinates, n_weights_per_side=8)

    x = {
        'alpha'              : alpha,
        'Re'                 : Re,
        **{f"kulfan_lower_{i}": kulfan_params["lower_weights"][i] for i in range(8)},
        **{f"kulfan_upper_{i}": kulfan_params["upper_weights"][i] for i in range(8)},
        "kulfan_TE_thickness": kulfan_params["TE_thickness"],
        "kulfan_LE_weight"   : kulfan_params["leading_edge_weight"]
    }

    log10_Re = np.log10(x["Re"])

    switch = (log10_Re - p["log10_Re_switch"]) / p["log10_Re_scale"]

    thickness_modes = [
        x[f"kulfan_upper_{i}"] - x[f"kulfan_lower_{i}"]
        for i in range(8)
    ]
    camber_modes = [
        x[f"kulfan_upper_{i}"] + x[f"kulfan_lower_{i}"]
        for i in range(8)
    ]

    CL0_high = sum([
        p[f"CL0_high_{i}"] * camber_modes[i]
        for i in range(8)
    ]) + p["CL0_high_LE"] * x["kulfan_LE_weight"]
    CL0_low = sum([
        p[f"CL0_low_{i}"] * camber_modes[i]
        for i in range(8)
    ]) + p["CL0_low_LE"] * x["kulfan_LE_weight"]

    CLa_high = p["CLa_high"] + p["delta_CLa_high_area"] * sum(thickness_modes)

    CLa_low = p["CLa_low"] + p["delta_CLa_low_area"] * sum(thickness_modes)

    CL0 = np.blend(
        switch,
        CL0_high,
        CL0_low
    )
    CLa = np.blend(
        switch,
        CLa_high,
        CLa_low
    )

    return CL0 + CLa * np.radians(x["alpha"])


if __name__ == '__main__':
    af = asb.Airfoil("ag36")
    alpha = np.linspace(-10, 10, 100)
    Re = 1e6
    CL = get_CL(af, alpha, Re)



import aerosandbox as asb
import aerosandbox.numpy as np
from typing import Union, Dict, Set, List
from pathlib import Path
from neuralfoil.gen2_architecture._basic_data_type import Data

npz_file_directory = Path(__file__).parent / "nn_weights_and_biases"

bl_x_points = Data.bl_x_points

def _sigmoid(x):
    return 1 / (1 + np.exp(-x))

def _inverse_sigmoid(x):
    return -np.log(1 / x - 1)


def get_aero_from_kulfan_parameters(
        kulfan_parameters: Dict[str, Union[float, np.ndarray]],
        alpha: Union[float, np.ndarray],
        Re: Union[float, np.ndarray],
        n_crit: Union[float, np.ndarray] = 9.0,
        xtr_upper: Union[float, np.ndarray] = 1.0,
        xtr_lower: Union[float, np.ndarray] = 1.0,
        model_size="large"
) -> Dict[str, Union[float, np.ndarray]]:

    ### Load the neural network parameters
    filename = npz_file_directory / f"nn-{model_size}.npz"
    if not filename.exists():
        raise FileNotFoundError(
            f"Could not find the neural network file {filename}, which contains the weights and biases.")

    data: Dict[str, np.ndarray] = np.load(filename)

    ### Prepare the inputs for the neural network
    input_rows: List[Union[float, np.ndarray]] = [
        *[kulfan_parameters["upper_weights"][i] for i in range(8)],
        *[kulfan_parameters["lower_weights"][i] for i in range(8)],
        kulfan_parameters["leading_edge_weight"],
        kulfan_parameters["TE_thickness"] * 50,
        np.sind(2 * alpha),
        np.cosd(alpha),
        1 - np.cosd(alpha) ** 2,
        (np.log(Re) - 12.5) / 3.5,
        # No mach
        (n_crit - 9) / 4.5,
        xtr_upper,
        xtr_lower,
    ]

    ### Handle the vectorization, where here we figure out how many cases the user wants to run
    N_cases = 1  # TODO rework this with np.atleast1d
    for row in input_rows:
        if np.length(row) > 1:
            if N_cases == 1:
                N_cases = np.length(row)
            else:
                if np.length(row) != N_cases:
                    raise ValueError(
                        f"The inputs to the neural network must all have the same length. (Conflicting lengths: {N_cases} and {np.length(row)})"
                    )

    for i, row in enumerate(input_rows):
        input_rows[i] = np.ones(N_cases) * row

    x = np.stack(input_rows, axis=1)  # N_cases x N_inputs
    ##### Evaluate the neural network

    ### First, determine what the structure of the neural network is (i.e., how many layers it has) by looking at the keys.
    # These keys come from the dictionary of saved weights/biases for the specified neural network.
    try:
        layer_indices: Set[int] = set([
            int(key.split(".")[1])
            for key in data.keys()
        ])
    except TypeError:
        raise ValueError(
            f"Got an unexpected neural network file format.\n"
            f"Dictionary keys should be strings of the form 'net.0.weight', 'net.0.bias', 'net.2.weight', etc.'.\n"
            f"Instead, got keys of the form {data.keys()}.\n"
        )
    layer_indices: List[int] = sorted(list(layer_indices))

    ### Now, set up evaluation of the basic neural network.
    def net(x: np.ndarray):
        """
        Evaluates the raw network (taking in scaled inputs and returning scaled outputs).
        Input `x` dims: N_cases x N_inputs
        Output `y` dims: N_cases x N_outputs
        """
        x = np.transpose(x)
        layer_indices_to_iterate = layer_indices.copy()

        while len(layer_indices_to_iterate) != 0:
            i = layer_indices_to_iterate.pop(0)
            w = data[f"net.{i}.weight"]
            b = data[f"net.{i}.bias"]
            x = w @ x + np.reshape(b, (-1, 1))

            if len(layer_indices_to_iterate) != 0:  # Don't apply the activation function on the last layer
                x = np.tanh(x)
        x = np.transpose(x)
        return x

    y = net(x)  # N_outputs x N_cases

    ### Then, flip the inputs and evaluate the network again.
    # The goal here is to embed the invariant of "symmetry across alpha" into the network evaluation.
    # (This was also performed during training, so the network is "intended" to be evaluated this way.)

    x_flipped = x + 0.  # This is a array-api-agnostic way to force a memory copy of the array to be made.
    x_flipped[:, :8] = x[:, 8:16] * -1  # switch kulfan_lower with a flipped kulfan_upper
    x_flipped[:, 8:16] = x[:, :8] * -1  # switch kulfan_upper with a flipped kulfan_lower
    x_flipped[:, 16] *= -1  # flip kulfan_LE_weight
    x_flipped[:, 18] *= -1  # flip sin(2a)
    x_flipped[:, 23] = x[:, 24]  # flip xtr_upper with xtr_lower
    x_flipped[:, 24] = x[:, 23]  # flip xtr_lower with xtr_upper

    y_flipped = net(x_flipped)

    ### The resulting outputs will also be flipped, so we need to flip them back to their normal orientation
    y_unflipped = y_flipped + 0.  # This is a array-api-agnostic way to force a memory copy of the array to be made.
    y_unflipped[:, 1] *= -1  # CL
    y_unflipped[:, 3] *= -1  # CM
    y_unflipped[:, 4] = y_flipped[:, 5]  # switch Top_Xtr with Bot_Xtr
    y_unflipped[:, 5] = y_flipped[:, 4]  # switch Bot_Xtr with Top_Xtr

    # switch upper and lower Ret, H
    y_unflipped[:, 6:6 + 32 * 2] = y_flipped[:, 6 + 32 * 3: 6 + 32 * 5]
    y_unflipped[:, 6 + 32 * 3: 6 + 32 * 5] = y_flipped[:, 6:6 + 32 * 2]

    # switch upper_bl_ue/vinf with lower_bl_ue/vinf
    y_unflipped[:, 6 + 32 * 2: 6 + 32 * 3] = -1 * y_flipped[:, 6 + 32 * 5: 6 + 32 * 6]
    y_unflipped[:, 6 + 32 * 5: 6 + 32 * 6] = -1 * y_flipped[:, 6 + 32 * 2: 6 + 32 * 3]

    ### Then, average the two outputs to get the "symmetric" result
    y_fused = (y + y_unflipped) / 2
    y_fused[:, 0] = _sigmoid(y_fused[:, 0])  # Analysis confidence, a binary variable

    # Unpack the neural network outputs
    # sqrt_Rex_approx = [((Data.bl_x_points[i] + 1e-2) / Re) ** 0.5 for i in range(Data.N)]
    # TODO

    analysis_confidence = y_fused[:, 0]
    CL = y_fused[:, 1] / 2
    CD = np.exp((y_fused[:, 2] - 2) * 2)
    CM = y_fused[:, 3] / 20
    Top_Xtr = y_fused[:, 4]
    Bot_Xtr = y_fused[:, 5]

    upper_bl_ue_over_vinf = y_fused[:, 6 + Data.N * 2:6 + Data.N * 3]
    lower_bl_ue_over_vinf = y_fused[:, 6 + Data.N * 5:6 + Data.N * 6]

    upper_theta = (
                          (10 ** y_fused[:, 6: 6 + Data.N]) - 0.1
    ) / (np.abs(upper_bl_ue_over_vinf) * np.reshape(Re, (-1, 1)))
    upper_H = 2.6 * np.exp(y_fused[:, 6 + Data.N: 6 + Data.N * 2])

    lower_theta = (
                            (10 ** y_fused[:, 6 + Data.N * 3: 6 + Data.N * 4]) - 0.1
    ) / (np.abs(lower_bl_ue_over_vinf) * np.reshape(Re, (-1, 1)))
    lower_H = 2.6 * np.exp(y_fused[:, 6 + Data.N * 4: 6 + Data.N * 5])


    results = {
        "analysis_confidence": analysis_confidence,
        "CL"                 : CL,
        "CD"                 : CD,
        "CM"                 : CM,
        "Top_Xtr"            : Top_Xtr,
        "Bot_Xtr"            : Bot_Xtr,
        **{
            f"upper_bl_theta_{i}": upper_theta[:, i]
            for i in range(Data.N)
        },
        **{
            f"upper_bl_H_{i}": upper_H[:, i]
            for i in range(Data.N)
        },
        **{
            f"upper_bl_ue/vinf_{i}": upper_bl_ue_over_vinf[:, i]
            for i in range(Data.N)
        },
        **{
            f"lower_bl_theta_{i}": lower_theta[:, i]
            for i in range(Data.N)
        },
        **{
            f"lower_bl_H_{i}": lower_H[:, i]
            for i in range(Data.N)
        },
        **{
            f"lower_bl_ue/vinf_{i}": lower_bl_ue_over_vinf[:, i]
            for i in range(Data.N)
        },
    }
    return {key: np.reshape(value, -1) for key, value in results.items()}


def get_aero_from_airfoil(
        airfoil: asb.Airfoil,
        alpha: Union[float, np.ndarray],
        Re: Union[float, np.ndarray],
        n_crit: Union[float, np.ndarray] = 9.0,
        xtr_upper: Union[float, np.ndarray] = 1.0,
        xtr_lower: Union[float, np.ndarray] = 1.0,
        model_size="large",
) -> Dict[str, Union[float, np.ndarray]]:

    airfoil_normalization = airfoil.normalize(return_dict=True)

    from aerosandbox.geometry.airfoil.airfoil_families import get_kulfan_parameters

    kulfan_parameters = get_kulfan_parameters(
        airfoil_normalization["airfoil"].coordinates,
        n_weights_per_side=8
    )

    return get_aero_from_kulfan_parameters(
        kulfan_parameters=kulfan_parameters,
        alpha=alpha + airfoil_normalization["rotation_angle"],
        Re=Re / airfoil_normalization["scale_factor"],
        n_crit=n_crit,
        xtr_upper=xtr_upper,
        xtr_lower=xtr_lower,
        model_size=model_size
    )


def get_aero_from_coordinates(
        coordinates: np.ndarray,
        alpha: Union[float, np.ndarray],
        Re: Union[float, np.ndarray],
        n_crit: Union[float, np.ndarray] = 9.0,
        xtr_upper: Union[float, np.ndarray] = 1.0,
        xtr_lower: Union[float, np.ndarray] = 1.0,
        model_size="large",
):
    return get_aero_from_airfoil(
        airfoil=asb.Airfoil(
            coordinates=coordinates
        ),
        alpha=alpha,
        Re=Re,
        n_crit=n_crit,
        xtr_upper=xtr_upper,
        xtr_lower=xtr_lower,
        model_size=model_size
    )


def get_aero_from_dat_file(
        filename,
        alpha: Union[float, np.ndarray],
        Re: Union[float, np.ndarray],
        model_size="large",
):
    with open(filename, "r") as f:
        raw_text = f.readlines()

    from aerosandbox.geometry.airfoil.airfoil_families import get_coordinates_from_raw_dat
    return get_aero_from_coordinates(
        coordinates=get_coordinates_from_raw_dat(raw_text=raw_text),
        alpha=alpha,
        Re=Re,
        model_size=model_size
    )


if __name__ == '__main__':

    airfoil = asb.Airfoil("dae11").repanel().normalize()
    # airfoil = asb.Airfoil("naca0050")
    # airfoil = asb.Airfoil("naca0012").add_control_surface(10, hinge_point_x=0.5)

    alpha = np.linspace(-5, 15, 50)
    Re = 1e6

    aero = get_aero_from_airfoil(
        airfoil, 3, Re,
        model_size="xxxlarge"
    )

    aeros = {}

    model_sizes = ["xxxlarge"]

    for model_size in model_sizes:
        aeros[f"NF-{model_size}"] = get_aero_from_airfoil(
            airfoil=airfoil,
            alpha=alpha,
            Re=Re,
            model_size=model_size
        )

    if True:

        aeros["XFoil"] = asb.XFoil(
            airfoil=airfoil,
            Re=Re,
            max_iter=20,
            xfoil_repanel=True
        ).alpha(alpha)

    import matplotlib.pyplot as plt
    import aerosandbox.tools.pretty_plots as p

    fig, ax = plt.subplots(2, 2, figsize=(10, 8))
    for label, aero in aeros.items():
        if "xfoil" in label.lower():
            a = aero["alpha"]
            kwargs = dict(
                color="k"
            )
        else:
            a = alpha
            kwargs = dict(
                alpha=0.6,
            )

        ax[0, 0].plot(a, aero["CL"], **kwargs)
        ax[0, 1].plot(a, aero["CD"], label=label, **kwargs)
        ax[1, 0].plot(a, aero["CM"], **kwargs)
        ax[1, 1].plot(aero["CD"], aero["CL"], **kwargs)
    plt.sca(ax[0, 1])
    plt.legend()
    ax[0, 0].set_xlabel("$\\alpha$")
    ax[0, 0].set_ylabel("$C_L$")

    ax[0, 1].set_xlabel("$\\alpha$")
    ax[0, 1].set_ylabel("$C_D$")
    ax[0, 1].set_ylim(bottom=0)

    ax[1, 0].set_xlabel("$\\alpha$")
    ax[1, 0].set_ylabel("$C_M$")

    ax[1, 1].set_xlabel("$C_D$")
    ax[1, 1].set_ylabel("$C_L$")
    ax[1, 1].set_xlim(left=0)

    from aerosandbox.tools.string_formatting import eng_string

    plt.suptitle(f"\"{airfoil.name}\" Airfoil at $Re_c = \\mathrm{{{eng_string(Re)}}}$")

    p.show_plot()



from neuralfoil.gen2_5_architecture.main import (
    get_aero_from_kulfan_parameters,
    get_aero_from_airfoil,
    get_aero_from_coordinates,
    get_aero_from_dat_file,
    bl_x_points,
)

__version__ = "0.2.4"
